#!/usr/bin/env python

from distutils.core import setup

setup(
    name='mittmcts',
    version='0.2',
    description='Man in the Table Information Set Monte Carlo Tree Search '
                'library',
    author='Dan Bravender',
    author_email='dan@bravender.net',
    url='https://github.com/dbravender/mittmcts',
    packages=['mittmcts'],
)
